# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import abc
import collections
import configparser
import datetime
import glob
import logging
import os
import statistics
from typing import Any
from typing import Dict
from typing import List
from typing import TypeVar

import ijson
import telemetry_peak_analyzer

BackendType = TypeVar("BackendType", bound="AbstractBackend")


class AbstractBackend(abc.ABC):
    """Abstract backend."""

    def __init__(self, conf: configparser.ConfigParser, section_name: str) -> None:
        """
        Constructor.

        :param configparser.ConfigParser conf: the conf object
        :param str section_name: the name of the section
        """
        self._conf = conf
        self._section_name = section_name
        self._logger = logging.getLogger(__name__)
        self._logger.info("Loading backend '%s'", self.__class__.__name__)

    @abc.abstractmethod
    def stats(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Any]:
        """
        Create statistics.

        :param datetime.datetime start_date: the start of the time interval
        :param datetime.datetime end_date: the end of the time interval
        :param list[str] index: the index
        :param list[str] dimensions: the dimensions
        :param dict[str, list[str]] dimensions_values: the values of dimensions (if available)
        :rtype: dict[str, any]
        :return: statistics for each dimension combination
        """

    @abc.abstractmethod
    def group_by(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
    ) -> List[Dict[str, str]]:
        """
        Group by.

        :param datetime.datetime start_date: the start of the time interval
        :param datetime.datetime end_date: the end of the time interval
        :param list[str] index: the index
        :param list[str] dimensions: list of dimensions to group by
        :rtype: list[dict[str, str]]
        :return: buckets for each dimension combination
        """


class TwoIndexTwoDimensionBackend(AbstractBackend, abc.ABC):
    """Backend accepting an index with two fields and two dimensions."""

    @abc.abstractmethod
    def stats(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """
        Create statistics.

        :param datetime.datetime start_date: the start of the time interval
        :param datetime.datetime end_date: the end of the time interval
        :param list[str] index: the index
        :param list[str] dimensions: the dimensions
        :param dict[str, list[str]] dimensions_values: the values of dimensions (if available)
        :rtype: dict[str, dict[str, dict[str, float]]]
        :return: a dictionary like:
                {
                    "dimension_0_value": {
                        "dimension_1_value": {
                            "sub_count_avg": average of submissions
                            "sub_count_max": maximum number of submissions
                            "samp_count_avg": average of samples
                            "samp_count_max": maximum number of samples
                            "samp_sub_count_max": maximum ratio between samples and submissions
                        }
                        ...
                    }
                    ...
                }
        """

    @abc.abstractmethod
    def group_by(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
    ) -> List[Dict[str, str]]:
        """
        Group by.

        :param datetime.datetime start_date: the start of the time interval
        :param datetime.datetime end_date: the end of the time interval
        :param list[str] index: the index
        :param list[str] dimensions: the dimensions
        :rtype: list[dict[str, str]]
        :return: a list of dictionaries like:
            [
                {
                  "dimension_0": "benign",
                  "dimension_1": "ExcelMsDocFile",
                  "index_1": "0015cc85a17d707e00b9881a149c232d181ad451",
                  "additional_dimension_0": "3549",
                  "additional_dimension_1": "API",
                  "count": 61
                }
                ...
            ]
        """


class JsonBackend(TwoIndexTwoDimensionBackend):
    """Backend using JSON files."""

    def __init__(self, conf: configparser.ConfigParser, file_path_wildcard: str) -> None:
        """Constructor."""
        super(JsonBackend, self).__init__(conf, section_name="not_used")
        self._file_paths = []
        for name in glob.glob(file_path_wildcard):
            self._file_paths.append(os.path.abspath(name))
        self._logger.info("Loaded files:")
        for file_path in self._file_paths:
            self._logger.info("\t%s", file_path)

    def stats(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """Implement interface."""
        # we do several passes so to keep memory usage to a minimum
        # pass 1, let us get all the dates
        dates = set([])
        for file_path in self._file_paths:
            with open(file_path, "r") as f:
                for json_doc in ijson.items(f, "item"):
                    index_0 = telemetry_peak_analyzer.ms_to_datetime(json_doc[index[0]])
                    if start_date <= index_0 < end_date:
                        dates.add(index_0.date())

        # pass 2, for each date get the stats
        buckets = collections.defaultdict(lambda: collections.defaultdict(dict))
        all_dims_0 = set([])
        all_dims_1 = set([])
        for day_date in sorted(dates):
            sub_count = collections.defaultdict(lambda: collections.defaultdict(int))
            samp_set = collections.defaultdict(lambda: collections.defaultdict(set))
            dims_0 = set([])
            dims_1 = set([])
            for file_path in self._file_paths:
                with open(file_path, "r") as f:
                    for json_doc in ijson.items(f, "item"):
                        index_0 = telemetry_peak_analyzer.ms_to_datetime(
                            json_doc[index[0]]
                        ).date()
                        index_1 = json_doc[index[1]]
                        dimension_0 = json_doc[dimensions[0]]
                        dimension_1 = json_doc[dimensions[1]]
                        if index_0 == day_date:
                            sub_count[dimension_0][dimension_1] += 1
                            samp_set[dimension_0][dimension_1].add(index_1)
                            dims_0.add(dimension_0)
                            dims_1.add(dimension_1)
            for dim_0 in dims_0:
                for dim_1 in dims_1:
                    try:
                        samp_sub_count = sub_count[dim_0][dim_1] / len(samp_set[dim_0][dim_1])
                    except ZeroDivisionError:
                        samp_sub_count = 0
                    buckets[day_date][dim_0][dim_1] = {
                        "sub_count": sub_count[dim_0][dim_1],
                        "samp_count": len(samp_set[dim_0][dim_1]),
                        "samp_sub_count": samp_sub_count,
                    }
            all_dims_0.update(dims_0)
            all_dims_1.update(dims_1)
        ret = collections.defaultdict(dict)
        for dim_0 in all_dims_0:
            for dim_1 in all_dims_1:
                d_slice = [buckets[x].get(dim_0, {}).get(dim_1, {}) for x in dates]
                ret[dim_0][dim_1] = {
                    "sub_count_avg": statistics.mean([x.get("sub_count", 0) for x in d_slice]),
                    "sub_count_max": max([x.get("sub_count", 0) for x in d_slice]),
                    "samp_count_avg": statistics.mean([x.get("samp_count", 0) for x in d_slice]),
                    "samp_count_max": max([x.get("samp_count", 0) for x in d_slice]),
                    "samp_sub_count_max": max([x.get("samp_sub_count", 0) for x in d_slice]),
                }
        return ret

    def group_by(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
    ) -> List[Dict[str, str]]:
        """Implement interface."""
        counters = collections.Counter()
        dimensions = dimensions + [index[1]]
        for file_path in self._file_paths:
            with open(file_path, "r") as f:
                for json_doc in ijson.items(f, "item"):
                    index_0 = telemetry_peak_analyzer.ms_to_datetime(json_doc[index[0]])
                    if start_date <= index_0 < end_date:
                        counters[tuple([json_doc[dimension] for dimension in dimensions])] += 1
        ret = []
        for key, count in counters.items():
            value = {attr: key[idx] for idx, attr in enumerate(dimensions)}
            value["count"] = count
            ret.append(value)
        return ret



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import datetime
import json
import logging
import os
import resource
import sys
from typing import Dict
from typing import List
from typing import Union


def ms_to_datetime(milliseconds: int) -> datetime.datetime:
    """
    Convert a given amount of milliseconds to a datetime object.

    :param int milliseconds: number of milliseconds
    :rtype: datetime.datetime
    :return: the datetime object
    """
    return datetime.datetime.fromtimestamp(milliseconds / 1000)


def datetime_to_sec(timestamp: datetime.datetime) -> int:
    """
    Convert a given timestamp to seconds since the epoch.

    :param datetime.datetime timestamp: the datetime object
    :rtype: int
    """
    return int((timestamp - datetime.datetime.utcfromtimestamp(0)).total_seconds())


def datetime_to_ms(timestamp: datetime.datetime) -> int:
    """
    Convert a given timestamp to milliseconds since the epoch.

    :param datetime.datetime timestamp: the datetime object
    :rtype: int
    """
    return datetime_to_sec(timestamp) * 1000


def save_to_json(obj: Union[Dict, List], file_path: str) -> None:
    """
    Save python object.

    :param dict|list obj: the python object
    :param str file_path: the path where to save to
    """
    with open(file_path, "w") as f:
        json.dump(obj, f, indent=2 * " ", sort_keys=True, default=str)


class MemoryFootprintFormatter(logging.Formatter):
    """Special formatter keeping track how much memory is used."""

    _DEFAULT_FMT = "%(levelname)s -> [%(asctime)s] %(message)s"
    _DEFAULT_DATEFMT = "%Y-%m-%d %H:%M:%S"

    def __init__(
        self,
        fmt: str = _DEFAULT_FMT,
        datefmt: str = _DEFAULT_DATEFMT,
    ):
        """Override method."""
        super(MemoryFootprintFormatter, self).__init__(fmt, datefmt)

    @classmethod
    def configure_logging(cls, level: int) -> None:
        """
        Configure some sane defaults.

        :param int level: the debugging level.
        """
        handler = logging.StreamHandler()
        formatter = cls()
        handler.setFormatter(formatter)
        logging.root.addHandler(handler)
        logging.root.setLevel(level)

    @staticmethod
    def _read_procfs_memory(procfs_mem_key: str) -> float:
        """
        Read memory usage from the /proc filesystem.

        :param str procfs_mem_key: the key to search
        :rtype: float
        :return: the value of the related key
        """
        procfs_fn = os.path.join("/", "proc", str(os.getpid()), "status")
        procfs_stats_scale = {
            "kB": 1024.0,
            "mB": 1024.0 * 1024.0,
            "KB": 1024.0,
            "MB": 1024.0 * 1024.0,
        }
        with open(procfs_fn) as pf:
            pf_data = pf.read()
            # get VmKey line e.g. "VmRSS:  9999  kB\n ..."
            i = pf_data.index(procfs_mem_key)
            # remove white-spaces
            v = pf_data[i:].split(None, 3)
            if len(v) < 3:
                return 0.0
            # scale to the unit that was asked for
            return float(v[1]) * procfs_stats_scale[v[2]]

    @staticmethod
    def _get_memory_size() -> float:
        """
        Retrieve total memory usage in bytes from /proc filesystem.

        :rtype: float
        :return: the value of the memory size
        """
        return MemoryFootprintFormatter._read_procfs_memory("VmSize:")

    @staticmethod
    def _get_resident_memory_size() -> float:
        """
        Retrieve resident memory usage in bytes from /proc filesystem.

        :rtype: float
        :return: the value of the memory size
        """
        return MemoryFootprintFormatter._read_procfs_memory("VmRSS:")

    @classmethod
    def _get_memory_consumption(cls) -> float:
        """
        Get the current memory consumption in megabytes.

        :rtype: float
        :return: the memory consumption
        """
        try:
            # We get the actual memory footprint
            res_mem_size = MemoryFootprintFormatter._get_resident_memory_size()
        except IOError:
            # On OSX we do not have /proc so we fallback to resource
            if sys.platform == "darwin":
                # IMPORTANT: on OSX results are given in Bytes while on Linux are given in KBytes
                res_mem_size = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            else:
                # If we are not on Linux or on OSX just give up
                res_mem_size = 1024.0 * 1024.0
        res_mem_size /= 1024.0 * 1024.0
        return res_mem_size

    def format(self, record: logging.LogRecord) -> str:
        """Override."""
        # A record can be formatted multiple times by SimpleLogger.
        if not hasattr(record, "formatted"):
            record.msg = "[%04dmb] %s" % (self._get_memory_consumption(), record.msg)
            setattr(record, "formatted", True)
        return super(MemoryFootprintFormatter, self).format(record)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import abc
import collections
import configparser
import datetime
import json
import logging
import statistics
from abc import ABC
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

import telemetry_peak_analyzer
from telemetry_peak_analyzer import backends
from telemetry_peak_analyzer import models


class AbstractAnalyzer(abc.ABC):
    """Abstract analyzer."""

    DEFAULT_GLOBAL_TABLE_AGE = datetime.timedelta(days=7)

    @staticmethod
    def _get_window_count(start_ts: int, end_ts: int) -> int:
        """
        Get the window count.

        :param int start_ts: milliseconds
        :param int end_ts: milliseconds
        :rtype: int
        :return: the window count
        """
        delta_hours = (end_ts - start_ts) // 1000 // 60 // 60
        window_count = delta_hours // 24 if delta_hours % 24 == 0 else delta_hours
        # if start_ts and end_ts from the buckets are the same, then count it as one window
        if window_count == 0:
            window_count = 1
        return window_count

    def __init__(
        self,
        conf: configparser.ConfigParser,
        backend: backends.BackendType,
        index: List[str],
        dimensions: List[str],
        start_ts: datetime.datetime,
        end_ts: datetime.datetime,
    ) -> None:
        """
        Constructor.

        :param configparser.ConfigParser conf: the conf object
        :param BackendType backend: the backend
        :param list[str] index: the index
        :param list[str] dimensions: the dimensions
        :param datetime.datetime start_ts: the beginning of the time interval
        :param datetime.datetime end_ts: the end of the time interval
        """
        self._conf = conf
        self._backend = backend
        self._index = index
        self._dimensions = dimensions
        self._start_ts = start_ts
        self._end_ts = end_ts
        self._logger = logging.getLogger(__name__)
        self._logger.info(
            "Loading analyzer '%s' with backend '%s'",
            self.__class__.__name__,
            self._backend.__class__.__name__,
        )

    @abc.abstractmethod
    def get_peaks(
        self,
        global_tables: Dict,
        local_tables: Dict,
        threshold: Optional[int] = None,
    ) -> Dict:
        """
        Get the peaks.

        :param dict global_tables: the global tables
        :param dict local_tables: the local tables
        :param int|None threshold: optional threshold
        :rtype: dict
        :return: the telemetry peaks
        """

    @abc.abstractmethod
    def refresh_global_tables(self, global_tables: Dict, local_tables: Dict) -> Dict:
        """
        Refresh the global tables.

        :param dict global_tables: the global tables
        :param dict local_tables: the local tables
        :rtype: dict
        :return: the global tables
        """

    @abc.abstractmethod
    def get_global_tables_stats(
        self,
        global_tables: Dict,
        threshold: Optional[int] = None,
    ) -> Dict:
        """
        Get statistics from the global table.

        :param dict global_tables: the global tables
        :param int|None threshold: optional threshold
        :rtype: dict
        :return: the global tables stats
        """

    @abc.abstractmethod
    def get_global_tables(self) -> Dict:
        """
        Get the global tables from our backend.

        :rtype: dict
        :return: global tables
        """

    @abc.abstractmethod
    def get_global_tables_from_file(self, file_path: str) -> Dict:
        """
        Load global tables from a file.

        :param str file_path: the file path where the load from
        :rtype: dict
        :return: the loaded object
        """

    @abc.abstractmethod
    def get_local_tables_stats(self, local_tables: Dict) -> Dict:
        """
        Get statistics from the local table.

        :param dict local_tables: local tables
        :rtype: dict
        :return: the local tables stats
        """

    @abc.abstractmethod
    def get_local_tables(self) -> Dict:
        """
        Get the local tables from our backend.

        :rtype: dict
        :return: the local tables
        """


class TwoIndexTwoDimensionAnalyzer(AbstractAnalyzer, ABC):
    """Analyzer using index and dimension with cardinality set to two."""

    LOCAL_MIN_COUNT = 50

    PEAK_GLOBAL_COUNT_WEIGHT = 0.8

    PEAK_MIN_SAMP_SUB_RATIO = 0.5

    CROSS_DIMENSIONS = []

    DIMENSIONS_METADATA = {}

    def _get_dimension_threshold(self, dimension: str, value: str) -> int:
        """Get the threshold for a given dimension and value, if available, 0 otherwise."""
        try:
            return self.DIMENSIONS_METADATA[dimension]["threshold"][value]
        except KeyError:
            return 0

    def _get_dimension_values(self, dimension: str) -> List[str]:
        """Get all values that a dimension can exhibit, if available."""
        try:
            return self.DIMENSIONS_METADATA[dimension]["values"]
        except KeyError:
            return []

    def _is_peak(
        self,
        local_table_stats: models.LocalTableStats,
        global_table_stats: models.GlobalTableStats,
    ) -> bool:
        """
        Return whether the peak is valid.

        :param LocalTableStats local_table_stats: stats about the local table
        :param GlobalTableStats global_table_stats: stats about the global table
        :rtype: bool
        :return: whether the peak is valid
        """
        if local_table_stats.sub_count < global_table_stats.threshold:
            return False
        temp = self.PEAK_GLOBAL_COUNT_WEIGHT * global_table_stats.samp_sub_count_max
        is_susp_overall_sub_samp_r = local_table_stats.samp_sub_count_mean > temp
        is_dominant_samp_sub = local_table_stats.samp_sub_ratio > self.PEAK_MIN_SAMP_SUB_RATIO
        temp = local_table_stats.samp_sub_count_mean + local_table_stats.samp_sub_count_std
        is_susp_samp_sub_var = local_table_stats.samp_sub_count_max > temp
        return is_susp_overall_sub_samp_r or is_dominant_samp_sub or is_susp_samp_sub_var

    def _update_global_table(
        self,
        global_table: models.GlobalTable,
        local_table: Dict[str, Dict[str, int]],
    ) -> models.GlobalTable:
        """
        Update the global table

        :param GlobalTable global_table: the global table
        :param dict[str, dict[str, int]] local_table: the local table
        :rtype: GlobalTable
        :return: a refreshed global table
        """
        window_count = global_table.window_count + 1
        sub_count = sum(local_table[self._index[1]].values())
        samp_count = len(local_table[self._index[1]])
        try:
            samp_sub_count = int(round(sub_count / samp_count))
        except ZeroDivisionError:
            samp_sub_count = 0
        sub_count_avg = int(
            round(
                (global_table.sub_count_avg * global_table.window_count + sub_count)
                / window_count
            )
        )
        samp_count_avg = int(
            round(
                (global_table.samp_count_avg * global_table.window_count + samp_count)
                / window_count
            )
        )
        samp_sub_count_avg = int(
            round(
                (global_table.samp_sub_count_avg * global_table.window_count + samp_sub_count)
                / window_count
            )
        )
        return models.GlobalTable(
            start_ts=global_table.start_ts,
            end_ts=self._end_ts,
            window_count=window_count,
            sub_count_avg=sub_count_avg,
            sub_count_max=max(global_table.sub_count_max, sub_count),
            samp_count_avg=samp_count_avg,
            samp_count_max=max(global_table.samp_count_max, samp_count),
            samp_sub_count_avg=samp_sub_count_avg,
            samp_sub_count_max=max(global_table.samp_sub_count_max, samp_sub_count),
            threshold_suggested=max(global_table.threshold_suggested, sub_count_avg),
        )

    def _infer_global_table(
        self,
        local_table: Dict[str, Dict[str, int]],
        threshold: int,
    ) -> models.GlobalTable:
        """
        Infer the global table from the local table.

        :param dict[str, dict[str, int]] local_table: the local table
        :param int threshold: the threshold to suggest
        :rtype: GlobalTable
        :return: the inferred global table
        """
        sub_count = sum(local_table[self._index[1]].values())
        samp_count = len(local_table[self._index[1]])
        try:
            samp_sub_count = int(round(sub_count / samp_count))
        except ZeroDivisionError:
            samp_sub_count = 0
        return models.GlobalTable(
            start_ts=self._start_ts,
            end_ts=self._end_ts,
            window_count=1,
            sub_count_avg=sub_count,
            sub_count_max=sub_count,
            samp_count_avg=samp_count,
            samp_count_max=samp_count,
            samp_sub_count_avg=samp_sub_count,
            samp_sub_count_max=samp_sub_count,
            threshold_suggested=max(sub_count, threshold),
        )

    def get_peaks(
        self,
        global_tables: Dict[str, Dict[str, models.GlobalTable]],
        local_tables: Dict[str, Dict[str, Dict[str, Dict[str, int]]]],
        threshold: Optional[int] = None,
    ) -> Dict[str, Dict[str, models.TelemetryPeak]]:
        """
        Get the peaks.

        :param dict[str, dict[str, GlobalTable]] global_tables: the global tables
        :param dict[str, dict[str, dict[str, dict[str, int]]]] local_tables: the local tables
        :param int|None threshold: optional threshold
        :rtype: dict[str, dict[str, TelemetryPeak]]
        :return: the telemetry peaks
        """
        local_tables_stats = self.get_local_tables_stats(local_tables)
        global_tables_stats = self.get_global_tables_stats(global_tables, threshold=threshold)
        peaks = collections.defaultdict(dict)
        for dimension_0 in local_tables_stats:
            for dimension_1 in local_tables_stats[dimension_0]:
                try:
                    local_table_stats = local_tables_stats[dimension_0][dimension_1]
                    global_table_stats = global_tables_stats[dimension_0][dimension_1]
                except KeyError:
                    continue
                if self._is_peak(local_table_stats, global_table_stats):
                    peaks[dimension_0][dimension_1] = models.TelemetryPeak(
                        sub_count=local_table_stats.sub_count,
                        samp_count=local_table_stats.samp_count,
                        samp_sub_count_max=local_table_stats.samp_sub_count_max,
                        samp_sub_count_mean=local_table_stats.samp_sub_count_mean,
                        samp_sub_count_std=local_table_stats.samp_sub_count_std,
                        samp_sub_ratio=local_table_stats.samp_sub_ratio,
                        global_samp_sub_count_max=global_table_stats.samp_sub_count_max,
                        global_threshold_suggested=global_table_stats.threshold,
                    )
        return peaks

    def refresh_global_tables(
        self,
        global_tables: Dict[str, Dict[str, models.GlobalTable]],
        local_tables: Dict[str, Dict[str, Dict[str, Dict[str, int]]]],
    ) -> Dict[str, Dict[str, models.GlobalTable]]:
        """
        Refresh the global tables.

        :param dict[str, dict[str, GlobalTable]] global_tables: the global tables
        :param dict[str, dict[str, dict[str, dict[str, int]]]] local_tables: the local tables
        :rtype: dict[str, dict[str, GlobalTable]]
        :return: the global tables
        """

        def get_dimensions_0() -> Set[str]:
            """Get the first dimension from all the global and local tables."""
            return set(global_tables.keys()).union(local_tables.keys())

        def get_dimensions_1(dim_0: str) -> Set[str]:
            """Get the second dimension from all the global and local tables."""
            return set(global_tables.get(dim_0, {}).keys()).union(
                local_tables.get(dim_0, {}).keys()
            )

        refreshed_global_tables = collections.defaultdict(dict)
        for dimension_0 in get_dimensions_0():
            for dimension_1 in get_dimensions_1(dimension_0):
                global_table = global_tables.get(dimension_0, {}).get(dimension_1, None)
                local_table = local_tables.get(dimension_0, {}).get(dimension_1, None)
                # If we have both tables for all dimensions, check their age
                if global_table and local_table:
                    # if the global table is more recent than the current time interval
                    if self._start_ts < global_table.end_ts:
                        new_table = global_table
                    # otherwise update the global table
                    else:
                        new_table = self._update_global_table(global_table, local_table)
                # if we only have the global table, then use the global table
                elif global_table and not local_table:
                    new_table = global_table
                # and if we only have the local table, infer a global table
                elif not global_table and local_table:
                    threshold = self._get_dimension_threshold(self._dimensions[0], dimension_0)
                    new_table = self._infer_global_table(local_table, threshold)
                # otherwise just skip
                else:
                    continue
                refreshed_global_tables[dimension_0][dimension_1] = new_table
        return refreshed_global_tables

    def get_global_tables_stats(
        self,
        global_tables: Dict[str, Dict[str, models.GlobalTable]],
        threshold: Optional[int] = None,
    ) -> Dict[str, Dict[str, models.GlobalTableStats]]:
        """
        Get statistics from the global table.

        :param dict[str, dict[str, GlobalTable]] global_tables: the global tables
        :param int|None threshold: optional threshold
        :rtype: dict[str, dict[str, GlobalTableStats]]
        :return: the global tables stats
        """
        global_stats = collections.defaultdict(dict)
        for dimension_0 in global_tables:
            for dimension_1 in global_tables[dimension_0]:
                global_table = global_tables[dimension_0][dimension_1]
                global_stats[dimension_0][dimension_1] = models.GlobalTableStats(
                    threshold=threshold or global_table.threshold_suggested,
                    samp_sub_count_max=global_table.samp_sub_count_max,
                )
        return global_stats

    def get_global_tables_from_file(
        self,
        file_path: str,
    ) -> Dict[str, Dict[str, models.GlobalTable]]:
        """
        Load global tables from a file.

        :param str file_path: the file path where the load from
        :rtype: dict[str, dict[str, GlobalTable]]
        :return: the global tables
        """
        with open(file_path, "r") as f:
            json_data = json.load(f)
        for dimension_0 in json_data:
            for dimension_1 in json_data[dimension_0]:
                table = json_data[dimension_0][dimension_1]
                json_data[dimension_0][dimension_1] = models.GlobalTable(
                    start_ts=datetime.datetime.strptime(table[0][:19], "%Y-%m-%d %H:%M:%S"),
                    end_ts=datetime.datetime.strptime(table[1][:19], "%Y-%m-%d %H:%M:%S"),
                    window_count=table[2],
                    sub_count_avg=table[3],
                    sub_count_max=table[4],
                    samp_count_avg=table[5],
                    samp_count_max=table[6],
                    samp_sub_count_avg=table[7],
                    samp_sub_count_max=table[8],
                    threshold_suggested=table[9],
                )
        return json_data

    def get_global_tables(self) -> Dict[str, Dict[str, models.GlobalTable]]:
        """
        Get the global tables.

        :rtype: dict[str, dict[str, GlobalTable]]
        :return: the global tables
        """
        end_date = datetime.datetime.utcnow()
        start_date = end_date - self.DEFAULT_GLOBAL_TABLE_AGE
        ret = self._backend.stats(
            start_date=start_date,
            end_date=end_date,
            index=self._index,
            dimensions=self._dimensions,
            dimensions_values={x: self._get_dimension_values(x) for x in self._dimensions},
        )
        global_tables = collections.defaultdict(dict)
        for dimension_0 in ret:
            for dimension_1 in ret[dimension_0]:
                start_ts = telemetry_peak_analyzer.datetime_to_ms(start_date)
                end_ts = telemetry_peak_analyzer.datetime_to_ms(end_date)
                threshold = self._get_dimension_threshold(self._dimensions[0], dimension_0)
                sub_count_avg = ret[dimension_0][dimension_1]["sub_count_avg"]
                sub_count_max = ret[dimension_0][dimension_1]["sub_count_max"]
                samp_count_avg = ret[dimension_0][dimension_1]["samp_count_avg"]
                samp_count_max = ret[dimension_0][dimension_1]["samp_count_max"]
                samp_sub_count_max = ret[dimension_0][dimension_1]["samp_sub_count_max"]
                try:
                    samp_sub_count_avg = int(round(sub_count_avg / samp_count_avg))
                except ZeroDivisionError:
                    samp_sub_count_avg = 0
                global_tables[dimension_0][dimension_1] = models.GlobalTable(
                    start_ts=start_date,
                    end_ts=end_date,
                    window_count=self._get_window_count(start_ts, end_ts),
                    sub_count_avg=int(round(sub_count_avg)),
                    sub_count_max=int(sub_count_max),
                    samp_count_avg=int(round(samp_count_avg)),
                    samp_count_max=int(samp_count_max),
                    samp_sub_count_avg=samp_sub_count_avg,
                    samp_sub_count_max=samp_sub_count_max,
                    threshold_suggested=max(int(round(sub_count_avg)), threshold),
                )
        return global_tables

    def get_local_tables_stats(
        self, local_tables: Dict[str, Dict[str, Dict[str, Dict[str, int]]]]
    ) -> Dict[str, Dict[str, models.LocalTableStats]]:
        """
        Get statistics from the local table.

        :param dict[str, dict[str, dict[str, dict[str, int]]]] local_tables: local tables
        :rtype: dict[str, dict[str, LocalTableStats]]
        :return: some statistics for each dimension combination
        """
        local_stats = collections.defaultdict(dict)
        for dimension_0 in local_tables:
            for dimension_1 in local_tables[dimension_0]:
                local_table = local_tables[dimension_0][dimension_1]
                sub_count = sum(local_table[self._index[1]].values())
                samp_count = len(local_table[self._index[1]])
                samp_sub_count = local_table[self._index[1]].values()
                # fails if len < 1, one fails, both fail
                try:
                    samp_sub_count_mean = statistics.mean(samp_sub_count)
                    samp_sub_count_max = max(samp_sub_count)
                except (statistics.StatisticsError, ValueError):
                    samp_sub_count_mean = 0.0
                    samp_sub_count_max = 0
                # fails if len < 2
                try:
                    samp_sub_count_std = statistics.stdev(samp_sub_count)
                except statistics.StatisticsError:
                    samp_sub_count_std = 0.0
                if sub_count > self.LOCAL_MIN_COUNT:
                    samp_sub_ratio = round(samp_sub_count_max / sub_count, 2)
                else:
                    samp_sub_ratio = 0.0
                local_stats[dimension_0][dimension_1] = models.LocalTableStats(
                    sub_count=sub_count,
                    samp_count=samp_count,
                    samp_sub_count_max=samp_sub_count_max,
                    samp_sub_count_mean=samp_sub_count_mean,
                    samp_sub_count_std=samp_sub_count_std,
                    samp_sub_ratio=samp_sub_ratio,
                    cross_stats={x: local_table[x] for x in self.CROSS_DIMENSIONS},
                )
        return local_stats

    def get_local_tables(self) -> Dict[str, Dict[str, Dict[str, Dict[str, int]]]]:
        """
        Get the local tables from our backend.

        :rtype: dict[str, dict[str, dict[str, dict[str, int]]]]
        :return: the local tables
        """
        terms = self._dimensions + self.CROSS_DIMENSIONS + [self._index[1]]
        json_data = self._backend.group_by(
            start_date=self._start_ts,
            end_date=self._end_ts,
            index=self._index,
            dimensions=self._dimensions + self.CROSS_DIMENSIONS,
        )
        local_tables = collections.defaultdict(dict)
        for item in json_data:
            dimension_0 = item[self._dimensions[0]]
            dimension_1 = item[self._dimensions[1]]
            if dimension_1 not in local_tables[dimension_0]:
                local_tables[dimension_0][dimension_1] = {
                    term: collections.defaultdict(int) for term in terms
                }
            for term in terms:
                local_tables[dimension_0][dimension_1][term][item[term]] += item["count"]
        return local_tables


class FileTypePeakAnalyzer(TwoIndexTwoDimensionAnalyzer):
    """Analyzer using index and dimension to track file types."""

    CROSS_DIMENSIONS = [
        "source.user_id",
        "source.origin",
    ]

    DIMENSIONS_METADATA = {
        "task.severity": {
            "values": ["malicious", "benign"],
            "threshold": {
                "malicious": 90,
                "benign": 500,
            },
        }
    }

    _INDEX = [
        "utc_timestamp",
        "file.sha1",
    ]

    _DIMENSIONS = [
        "task.severity",
        "file.llfile_type",
    ]

    def __init__(
        self,
        conf: configparser.ConfigParser,
        backend: backends.BackendType,
        start_ts: datetime.datetime,
        end_ts: datetime.datetime,
    ) -> None:
        """
        Constructor.

        :param configparser.ConfigParser conf: the conf object
        :param backendType backend: the backend
        :param datetime.datetime start_ts: the beginning of the time interval
        :param datetime.datetime end_ts: the end of the time interval
        """
        if not isinstance(backend, backends.TwoIndexTwoDimensionBackend):
            raise ValueError("Backend is not compatible with the chosen analyzer")
        super(FileTypePeakAnalyzer, self).__init__(
            conf=conf,
            index=self._INDEX,
            dimensions=self._DIMENSIONS,
            backend=backend,
            start_ts=start_ts,
            end_ts=end_ts,
        )


class NetworkTypePeakAnalyzer(TwoIndexTwoDimensionAnalyzer):
    """Analyzer using index and dimension to track network peaks."""

    CROSS_DIMENSIONS = [
        "source.user_id",
    ]

    DIMENSIONS_METADATA = {
        "event.impact": {
           "values": ["70", "30"],
        }
    }

    _INDEX = [
        "utc_timestamp",
        "event.id",
    ]

    _DIMENSIONS = [
        "event.impact",
        "threat.name.keyword",
    ]

    def __init__(
        self,
        conf: configparser.ConfigParser,
        backend: backends.BackendType,
        start_ts: datetime.datetime,
        end_ts: datetime.datetime,
    ) -> None:
        """
        Constructor.

        :param configparser.ConfigParser conf: the conf object
        :param backendType backend: the backend
        :param datetime.datetime start_ts: the beginning of the time interval
        :param datetime.datetime end_ts: the end of the time interval
        """
        if not isinstance(backend, backends.TwoIndexTwoDimensionBackend):
            raise ValueError("Backend is not compatible with the chosen analyzer")
        # Relax the default age of a global table when not present
        self.DEFAULT_METRIC_TABLE_AGE = datetime.timedelta(days=3)
        super(NetworkTypePeakAnalyzer, self).__init__(
            conf=conf,
            index=self._INDEX,
            dimensions=self._DIMENSIONS,
            backend=backend,
            start_ts=start_ts,
            end_ts=end_ts,
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import collections
import configparser
import datetime
import logging
from typing import Any
from typing import Dict
from typing import List

from telemetry_peak_analyzer import backends

try:
    from tina_client.storage import readers
except ImportError:
    raise ImportError("The Tina backend requires tina-client.") from None


logging.getLogger("elasticsearch").setLevel(logging.WARNING)


class TinaBackend(backends.TwoIndexTwoDimensionBackend):
    """
    Backend using Tina (an internal Elasticsearch cluster).

    Note: in this class we hide all the specialization required to port the existing logic; note
        that this is unlikely to work with different analyzer and generalizing this bit further
        might not be worth the cost.
    """

    @staticmethod
    def _get_aggregation_query(
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict]:
        """
        Return a fully loaded Elasticsearch aggregation query.

        :param list[str] index: the index
        :param list[str] dimensions: the dimensions
        :param dict[str, list[str]] dimensions_values: the values of dimensions (if available)
        :rtype: dict[str, dict]
        :return: the aggregation to run
        """

        def _get_aggregation() -> Dict[str, Dict]:
            """Return the aggregation for each dimension."""
            return {
                "sub_per_day": {
                    "date_histogram": {
                        "field": index[0],
                        "calendar_interval": "day",
                        "min_doc_count": 0,
                    },
                    "aggs": {
                        "samp_count": {
                            "cardinality": {
                                "field": index[1],
                            }
                        },
                        "samp_sub_count": {
                            "bucket_script": {
                                "buckets_path": {
                                    "sub_count": "_count",
                                    "samp_count": "samp_count",
                                },
                                "script": "params.sub_count/params.samp_count",
                            }
                        },
                    },
                },
                "samp_count_avg": {"avg_bucket": {"buckets_path": "sub_per_day.samp_count"}},
                "samp_count_max": {"max_bucket": {"buckets_path": "sub_per_day.samp_count"}},
                "samp_sub_count_max": {
                    "max_bucket": {"buckets_path": "sub_per_day>samp_sub_count"}
                },
                "sub_count_avg": {"avg_bucket": {"buckets_path": "sub_per_day._count"}},
                "sub_count_max": {"max_bucket": {"buckets_path": "sub_per_day._count"}},
            }

        dimension_0_values = dimensions_values.get(dimensions[0], [])
        return {
            "aggs": {
                "my_buckets": {
                    "composite": {
                        "sources": [{"dimension": {"terms": {"field": dimensions[1]}}}]
                    },
                    "aggs": {
                        dimension_0_values[0]: {
                            "filter": {"term": {dimensions[0]: dimension_0_values[0]}},
                            "aggs": _get_aggregation(),
                        },
                        dimension_0_values[1]: {
                            "filter": {"term": {dimensions[0]: dimension_0_values[1]}},
                            "aggs": _get_aggregation(),
                        },
                    },
                }
            },
            "size": 0,
        }

    @staticmethod
    def _parse_aggregation_output(
        buckets: List[Dict[str, Dict[str, Any]]],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """
        Parse the aggregation from Elasticsearch into a backend-independent format.

        :param list[dict[str, dict[str, any]]] buckets: the aggregation buckets
        :param list[str] dimensions: the dimensions
        :param dict[str, list[str]] dimensions_values: the values of some dimensions (if available)
        :rtype: dict[str, dict[str, dict[str, float]]]
        :return: the aggregation results indexed by dimension
        """
        aggregation_keys = frozenset(
            [
                "sub_count_avg",
                "sub_count_max",
                "samp_count_avg",
                "samp_count_max",
                "samp_sub_count_max",
            ]
        )
        dimension_0_values = dimensions_values.get(dimensions[0], [])
        ret = collections.defaultdict(dict)
        for bucket in buckets:
            for dimension_0 in dimension_0_values:
                if bucket[dimension_0]["doc_count"] > 0:
                    dimension_1 = bucket["key"]["dimension"]
                    ret[dimension_0][dimension_1] = {
                        k: bucket[dimension_0][k]["value"]
                        for k in bucket[dimension_0]
                        if k in aggregation_keys
                    }
        return ret

    def __init__(self, conf: configparser.ConfigParser, section_name: str) -> None:
        """Constructor."""
        super(TinaBackend, self).__init__(conf, section_name)
        self._tina_reader = readers.BulkFileSubmissionReader(conf, section_name)
        self._logger.info(
            "Loading backend '%s' from section '%s'",
            self.__class__.__name__,
            section_name,
        )

    def stats(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """Implement interface."""
        query = self._get_aggregation_query(index, dimensions, dimensions_values)
        buckets = []
        while True:
            ret = self._tina_reader.search_raw(
                start_ts=start_date,
                end_ts=end_date,
                query=query,
                limit=0,
            )
            buckets.extend(ret["aggregations"]["my_buckets"]["buckets"])
            after_key = ret["aggregations"]["my_buckets"].get("after_key")
            if not after_key:
                break
            query["aggs"]["my_buckets"]["composite"]["after"] = after_key
        return self._parse_aggregation_output(buckets, dimensions, dimensions_values)

    def group_by(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
    ) -> List[Dict[str, str]]:
        """Implement interface."""
        buckets = self._tina_reader.aggregate(
            start_ts=start_date,
            end_ts=end_date,
            terms=dimensions + [index[1]],
            limit=None,
        )
        # This might or might not be a generator so in-place edit can't work
        new_buckets = []
        terms_set = set(dimensions + [index[1]])
        for bucket in buckets:
            for k in bucket.keys():
                if k in terms_set:
                    # Make sure we always deal with strings
                    bucket[k] = str(bucket[k])
            new_buckets.append(bucket)
        return new_buckets


class NetworkTinaBackend(TinaBackend):
    """New backend specialized for network events."""

    def __init__(self, conf: configparser.ConfigParser, section_name: str) -> None:
        """Constructor."""
        backends.TwoIndexTwoDimensionBackend.__init__(self, conf, section_name)
        self._tina_reader = readers.BulkEventReader(conf, section_name)
        self._logger.info(
            "Loading backend '%s' from section '%s'",
            self.__class__.__name__,
            section_name,
        )

    def stats(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
        dimensions_values: Dict[str, List[str]],
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        return super(NetworkTinaBackend, self).stats(start_date, end_date, index, dimensions, dimensions_values)

    def group_by(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        index: List[str],
        dimensions: List[str],
    ) -> List[Dict[str, str]]:
        return super(NetworkTinaBackend, self).group_by(start_date, end_date, index, dimensions)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import collections


TelemetryPeak = collections.namedtuple(
    "TelemetryPeak",
    [
        "sub_count",
        "samp_count",
        "samp_sub_count_max",
        "samp_sub_count_mean",
        "samp_sub_count_std",
        "samp_sub_ratio",
        "global_samp_sub_count_max",
        "global_threshold_suggested",
    ],
)


GlobalTable = collections.namedtuple(
    "GlobalTable",
    [
        "start_ts",
        "end_ts",
        "window_count",
        "sub_count_avg",
        "sub_count_max",
        "samp_count_avg",
        "samp_count_max",
        "samp_sub_count_avg",
        "samp_sub_count_max",
        "threshold_suggested",
    ],
)


LocalTableStats = collections.namedtuple(
    "LocalTableStats",
    [
        "sub_count",
        "samp_count",
        "samp_sub_count_max",
        "samp_sub_count_mean",
        "samp_sub_count_std",
        "samp_sub_ratio",
        "cross_stats",
    ],
)


GlobalTableStats = collections.namedtuple(
    "GlobalTableStats",
    [
        "samp_sub_count_max",
        "threshold",
    ],
)



#!/usr/bin/env python
# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
"""
See https://stackoverflow.com/questions/62983756/what-is-pyproject-toml-file-for
"""
import setuptools

if __name__ == "__main__":
    setuptools.setup()



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import argparse
import configparser
import datetime
import logging
import os
import sys
from typing import Optional
from typing import Tuple

import telemetry_peak_analyzer


def import_class(clazz_name: str) -> type:
    """
    Import the module and return the class.

    Example:
        > clazz = get_clazz_from_module("package.module.ClassName")
        > logging.debug("Instantiating %s instance...", clazz.__name__)
        > obj = clazz(conf)

    :param str clazz_name: class name in 'module.Class' form
    :rtype: type
    :return: the loadable type
    :raises ImportError: if the class name is not valid
    """
    if "." not in clazz_name:
        raise ImportError(f"Class '{clazz_name}' does not appear to be in module.Class form")

    try:
        only_clazz = clazz_name.split(".")[-1]
        only_module = ".".join(clazz_name.split(".")[:-1])
        mod = __import__(only_module, fromlist=[only_clazz])
        return getattr(mod, only_clazz)
    except AttributeError as ae:
        raise ImportError(f"Class '{clazz_name}' not found") from ae


def is_valid_date(date_str: str) -> datetime.date:
    """
    Validate a date and return a datetime object.

    :param str date_str: the datetime object as a string
    :rtype: datetime.date
    :return: the parsed date
    :raises ValueError: if the date is not valid
    """
    try:
        return datetime.datetime.strptime(date_str, "%Y-%m-%d").date()
    except ValueError:
        raise ValueError(f"Not a valid date: '{date_str}'") from None


def parse_date_options(
    start_ts: datetime.date,
    end_ts: datetime.date,
    delta: int,
    delay: int,
) -> Tuple[datetime.datetime, datetime.datetime]:
    """
    Validate the date options.

    :param datetime.date start_ts: the start of the time interval
    :param datetime.date end_ts: the end of the time interval
    :param int delta: the length of the time interval
    :param int delay: the delay of the time interval
    :rtype: tuple[datetime.datetime, datetime.datetime]
    :return: the validated datetime objects
    :raises ValueError: if the provided interval is not valid
    """
    if start_ts and end_ts:
        if end_ts <= start_ts:
            raise ValueError(f"Invalid time interval {start_ts} - {end_ts}")
    else:
        utc_now = datetime.datetime.utcnow()
        end_ts = (utc_now - datetime.timedelta(days=delay)).date()
        start_ts = end_ts - datetime.timedelta(days=delta)
    return (
        datetime.datetime.combine(start_ts, datetime.datetime.min.time()),
        datetime.datetime.combine(end_ts, datetime.datetime.min.time()),
    )


def run(
    config: configparser.ConfigParser,
    analyzer_class: type,
    backend_class: type,
    backend_input: str,
    start_ts: datetime.datetime,
    end_ts: datetime.datetime,
    threshold: int,
    global_table_path: str,
    output_file_path: Optional[str] = None,
) -> int:
    """Run the telemetry peak analyzer."""
    logger = logging.getLogger(__name__)
    logger.info("Loading Peak Analyzer from %s to %s with t=%s", start_ts, end_ts, threshold)
    peak_analyzer = analyzer_class(
        conf=config,
        backend=backend_class(config, backend_input),
        start_ts=start_ts,
        end_ts=end_ts,
    )
    try:
        logger.info("Loading global tables from file '%s'", global_table_path)
        global_tables = peak_analyzer.get_global_tables_from_file(global_table_path)
    except IOError as ioe:
        logger.info("\tFailed: %s", str(ioe))
        logger.info("Loading global tables from the backend")
        global_tables = peak_analyzer.get_global_tables()
    logger.info("Loading local tables")
    local_tables = peak_analyzer.get_local_tables()
    peaks = peak_analyzer.get_peaks(global_tables, local_tables, threshold=threshold)

    logger.info("Getting peaks")
    for dimension_0 in peaks:
        for dimension_1 in peaks[dimension_0]:
            logger.info("TelemetryPeak(%s, %s)", dimension_0, dimension_1)
            peak = peaks[dimension_0][dimension_1]
            for k, v in peak._asdict().items():
                logger.info("\t%s: %s", k, round(v, 2))
    if output_file_path:
        logger.info("Saving output to: %s", output_file_path)
        telemetry_peak_analyzer.save_to_json(peaks, output_file_path)

    logger.info("Refreshing global tables")
    global_tables = peak_analyzer.refresh_global_tables(global_tables, local_tables)

    logger.info("Saving global tables to '%s'", global_table_path)
    telemetry_peak_analyzer.save_to_json(global_tables, global_table_path)
    return 0


def parse_and_run_command():
    """
    Examples:
        # python -m telemetry_peak_analyzer \
            -b telemetry_peak_analyzer.backends.JsonBackend -n "~/data.*.json" \
            -s 2020-07-01 -e 2021-08-01 -t 10
        # python -m telemetry_peak_analyzer -c config.ini \
            -b telemetry_peak_analyzer.backends.tina.TinaBackend -n tina_backend \
            -s 2020-07-01 -e 2021-08-01 -t 10
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        default="./data/config.ini",
        type=str,
        help="read config from here",
    )
    # Time interval option 1: specify start and end datetime
    parser.add_argument(
        "-s",
        "--start-date",
        dest="start_date",
        default=None,
        type=is_valid_date,
        help="the start of the time interval in 'YYYY:mm:dd' format",
    )
    parser.add_argument(
        "-e",
        "--end-date",
        dest="end_date",
        default=None,
        type=is_valid_date,
        help="the end of the time interval in 'YYYY:mm:dd' format",
    )
    # Time interval option 2: specify the length and the delay of the time interval
    parser.add_argument(
        "-d",
        "--delta",
        dest="delta",
        default=1,
        type=int,
        help="the length of the time interval starting from now",
    )
    parser.add_argument(
        "-k",
        "--delay",
        dest="delay",
        default=0,
        type=int,
        help="the delay of the time interval expressed in days",
    )
    # Other options
    parser.add_argument(
        "-t",
        "--threshold",
        dest="threshold",
        default=None,
        type=int,
        help="the threshold used by the telemetry peak analyzer",
    )
    parser.add_argument(
        "-a",
        "--analyzer",
        dest="analyzer_class",
        default="telemetry_peak_analyzer.analyzers.FileTypePeakAnalyzer",
        type=import_class,
        help="the full class name of the analyzer used to process telemetry data",
    )
    parser.add_argument(
        "-b",
        "--backend-class",
        dest="backend_class",
        default="telemetry_peak_analyzer.backends.JsonBackend",
        type=import_class,
        help="the full class name of the backend used to read the telemetry from",
    )
    parser.add_argument(
        "-n",
        "--backend-input",
        dest="backend_input",
        required=True,
        type=str,
        help="the backend input, section name when reading remotely or a file for local input",
    )
    parser.add_argument(
        "-m",
        "--global-table",
        dest="global_table_path",
        default="global_table.json",
        type=str,
        help="the path to the global tables, either to load from, or to save to",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        dest="output_file",
        default=None,
        type=str,
        help="the path to output file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        default=False,
        action="store_true",
        help="whether to be verbose",
    )

    # Parse options and init the logger
    args = parser.parse_args()
    conf = configparser.ConfigParser()
    conf.read(args.config_file)
    start_date, end_date = parse_date_options(
        args.start_date,
        args.end_date,
        args.delta,
        args.delay,
    )
    log_level = logging.DEBUG if args.verbose else logging.INFO
    telemetry_peak_analyzer.MemoryFootprintFormatter.configure_logging(log_level)

    # Run
    return run(
        conf,
        args.analyzer_class,
        args.backend_class,
        args.backend_input,
        start_date,
        end_date,
        args.threshold,
        os.path.abspath(args.global_table_path),
        os.path.abspath(args.output_file) if args.output_file else None,
    )


if __name__ == "__main__":
    sys.exit(parse_and_run_command())
